{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "0dc32557",
   "metadata": {},
   "source": [
    "# *Tekstovni vmesnik\n",
    "\n",
    "**Opomba:** Do vključno študijskega leta 2021/22 je bil del predmeta Uvod v programiranje tudi projektna naloga, ki je obsegala pisanje spletne storitve. Poglavja o tem so ostala v zapiskih, vendar so označena z *, saj niso več del učnega programa.\n",
    "\n",
    "Za primer večjega programa v Pythonu bomo napisali enostaven program, ki nam bo omogočal beleženje opravil. Običajno je naš prvi instinkt, da poženemo urejevalnik kode in začnemo gledati, kako bi nastavili barvo gumbov. Vendar ta način običajno vodi v zmedeno kodo in posledično tudi neuporaben vmesnik, ki ga je zelo težko spreminjati. Zato se moramo stvari lotiti v pravem vrstnem redu in prvi korak je, da zapremo urejevalnik kode.\n",
    "\n",
    "## Skica vmesnika\n",
    "\n",
    "Nato na papir (ali na tablico) skiciramo, kaj pričakujemo od našega programa. Kakšen vmesnik bo ponujal, kam bodo izbire vodile uporabnika, … Ni treba, da je skica preveč natančna, važno je le, da na njej zajamemo vse, kar si v danem trenutku predstavljamo. Risanje je hitro in poceni, zato ni nič hudega, če s skico zgrešimo, saj jo vedno lahko popravimo. Primer skice, ki je nastala pred začetkom pisanja našega programa je:\n",
    "\n",
    "![](slike/skica-vmesnika.png)\n",
    "\n",
    "## Skica modela\n",
    "\n",
    "Tudi ko bo skica končana, se še ne bomo lotili pisanja vmesnika. Bolj pomembno je, da ugotovimo, kaj je naš _model_, torej predstavitev podatkov, s katerimi bomo delali. Če najprej model napišemo dobro, bo nudil trdno osnovo vmesniku (kot bomo videli celo več vmesnikom). Če pa se najprej lotimo vmesnika, bodo podatki zmešani z vmesnikom in ob prehodu na drug vmesnik (na primer iz tekstovnega na spletnega) bomo morali spremeniti veliko kode. Pri določitvi modela moramo ugotoviti, kaj so osnovni podatki, s katerimi želimo delati, in kakšne operacije bodo podpirali. Ker je Python objektni jezik, bomo vsako vrsto podatkov predstavili s svojim razredom. Pri določanju tega, katere atribute in metode mora vsebovati, nam bo v pomoč skica vmesnika.\n",
    "\n",
    "![](slike/skica-modela.png)\n",
    "\n",
    "Na skici označimo, katere razrede bomo imeli. Ker bomo beležili opravila, bomo vsekakor potrebovali razred `Opravilo`. Opravila bomo združevali v seznamih (domača opravila, službena opravila, …), katerih razred bomo poimenovali `Spisek`, da ne bo zmede s Pythonovimi seznami. Na koncu pa bomo potrebovali še krovni razred `Stanje`, ki bo zajemal vse podatke, ki jih potrebujemo za opis celotnega stanja. Vsak izmed razredov bo imel določene atribute, ki jih napišemo na vrhu, ter metode, ki jih napišemo na dnu. Povezav med razredi ne bomo pisali pod atribute, temveč jih bomo označili s povezavami na diagramu. Na vsaki povezavi si tudi zapišemo, koliko objektov je povezanih med seboj. Na primer, stanje ima lahko več spiskov, vendar le en aktualni spisek. Spisek ima eno stanje in več opravil.\n",
    "\n",
    "## Implementacija modela\n",
    "\n",
    "Na osnovi spiska zapišemo definicije ustreznih razredov. Najbolje, da kar v datoteko `model.py`. Ker bomo imeli opravka z roki opravil, si bomo pomagali z vgrajeno knjižnico [`datetime`](https://docs.python.org/3/library/datetime.html), ki med drugim ponuja razreda `date` za delo z datumi. Ko bomo implementirali metode, bomo opazili, da naša skica ni bila popolna (metod za popravljanje podatkov ne bomo rabili, na opravilih pa nam manjka metoda za ugotavljanje, ali je rok zamujen). Nič hudega, saj za skico nismo porabili preveč časa."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "d016246f",
   "metadata": {},
   "outputs": [],
   "source": [
    "from datetime import date\n",
    "\n",
    "\n",
    "class Stanje:\n",
    "    def __init__(self):\n",
    "        self.spiski = []\n",
    "        self.aktualni_spisek = None\n",
    "\n",
    "    def dodaj_spisek(self, spisek):\n",
    "        self.spiski.append(spisek)\n",
    "        if not self.aktualni_spisek:\n",
    "            self.aktualni_spisek = spisek\n",
    "\n",
    "    def pobrisi_spisek(self, spisek):\n",
    "        self.spiski.remove(spisek)\n",
    "\n",
    "    def zamenjaj_spisek(self, spisek):\n",
    "        self.aktualni_spisek = spisek\n",
    "\n",
    "    def dodaj_opravilo(self, opravilo):\n",
    "        self.aktualni_spisek.dodaj_opravilo(opravilo)\n",
    "\n",
    "    def pobrisi_opravilo(self, opravilo):\n",
    "        self.aktualni_spisek.pobrisi_opravilo(opravilo)\n",
    "\n",
    "\n",
    "class Spisek:\n",
    "    def __init__(self, ime):\n",
    "        self.ime = ime\n",
    "        self.opravila = []\n",
    "\n",
    "    def dodaj_opravilo(self, opravilo):\n",
    "        self.opravila.append(opravilo)\n",
    "\n",
    "    def stevilo_zamujenih(self):\n",
    "        stevilo = 0\n",
    "        for opravilo in self.opravila:\n",
    "            if opravilo.zamuja():\n",
    "                stevilo += 1\n",
    "        return stevilo\n",
    "\n",
    "    def stevilo_vseh(self):\n",
    "        return len(self.opravila)\n",
    "\n",
    "\n",
    "class Opravilo:\n",
    "    def __init__(self, ime, opis, rok, opravljeno=False):\n",
    "        self.ime = ime\n",
    "        self.opis = opis\n",
    "        self.rok = rok\n",
    "        self.opravljeno = opravljeno\n",
    "\n",
    "    def opravi(self):\n",
    "        self.opravljeno = True\n",
    "\n",
    "    def zamuja(self):\n",
    "        rok_pretekel = self.rok and self.rok < date.today()\n",
    "        return not self.opravljeno and rok_pretekel"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "592c73ce",
   "metadata": {},
   "source": [
    "## Implementacija vmesnika\n",
    "\n",
    "Ko je model v osnovi napisan, se lahko lotimo vmesnika. Za začetek se bomo lotili tekstovnega vmesnika, ki bo uporabniku izpisal trenutni prikaz ter s konzole prebral njegova navodila. Da bomo dobro poskrbeli za ločitev od modela, bomo vmesnik napisali v ločeno datoteko `tekstovni_vmesnik.py`. Paziti moramo, da iz vmesnika do modela dostopamo prek metod, saj bo tako celotno vedenje definirano v `model.py`. Prav tako je treba paziti, da iz modela nikoli ne dostopamo do vmesnika. To enostavno preprečimo tako, da se v `model.py` nikoli ne skličemo na datoteko `tekstovni_vmesnik.py`. Ker sta obe datoteki v istem imeniku, lahko model uvozimo kar kot:\n",
    "\n",
    "```python\n",
    "from model import Stanje\n",
    "\n",
    "stanje = Stanje()\n",
    "```\n",
    "\n",
    "Za začetek bo naše stanje prazno, kmalu pa si bomo pogledali, kako ga shranimo v datoteko in kasneje preberemo nazaj. Pisanja vmesnika se lotimo _od zgoraj navzdol_. Začnemo z osnovno zanko, ki bo pokazala pozdravno sporočilo, nato pa v nedogled uporabniku ponujala osnovni zaslon. Zato bo osnovna funkcija sledeča:\n",
    "\n",
    "```python\n",
    "def tekstovni_vmesnik():\n",
    "    prikazi_pozdravno_sporocilo()\n",
    "    while True:\n",
    "        osnovni_zaslon()\n",
    "```\n",
    "\n",
    "Osnovni zaslon bo najprej pokazal aktualna opravila, nato pa prebral ukaz in odvisno od njega poklical naslednjo funkcijo:\n",
    "\n",
    "```python\n",
    "DODAJ_OPRAVILO = 'dodaj'\n",
    "OPRAVI_OPRAVILO = 'opravi'\n",
    "\n",
    "def osnovni_zaslon():\n",
    "    prikazi_aktualna_opravila()\n",
    "    ukaz = preberi_ukaz()\n",
    "    if ukaz == DODAJ_OPRAVILO:\n",
    "        dodaj_opravilo()\n",
    "    elif ukaz == OPRAVI_OPRAVILO:\n",
    "        opravi_opravilo()\n",
    "    ...\n",
    "```\n",
    "\n",
    "Vsak ukaz bomo predstavili s svojim nizom. Da se ne bomo zatipkali in npr. v funkciji `osnovni_zaslon` pisali `'dodaj'`, v funkciji `preberi_ukaz` pa `'Dodaj'`, bomo ukaze shranili v konstante, ki so običajne spremenljivke, vendar s samimi velikimi črkami označujemo, da njihovih vrednosti ne bomo spreminjali. Če se bomo zatipkali pri pisanju imena konstante, nas bo na to opozoril urejevalnik.\n",
    "\n",
    "Funkcije bomo tako drobili naprej, na ustreznih točkah pa bomo poklicali metode na modelu. Na primer:\n",
    "\n",
    "```python\n",
    "def dodaj_opravilo():\n",
    "    ime = input('Ime opravila> ')\n",
    "    ...\n",
    "    opravilo = Opravilo(ime, ...)\n",
    "    stanje.aktualni_spisek.dodaj_opravilo(opravilo)\n",
    "```\n",
    "\n",
    "## Shranjevanje stanja\n",
    "\n",
    "Da ne bomo ob zaprtju programa izgubili vseh podatkov, si jih bomo pred zaprtjem shranili v datoteko, ki jo bomo ob vsakem zagonu prebrali. Tako bo naš vmesnik sledeče oblike:\n",
    "\n",
    "```python\n",
    "from stanje import Stanje\n",
    "\n",
    "IME_DATOTEKE = 'stanje.json'\n",
    "stanje = Stanje.preberi_iz_datoteke(IME_DATOTEKE)\n",
    "\n",
    "...\n",
    "\n",
    "def osnovni_zaslon():\n",
    "    while True:\n",
    "        ...\n",
    "        elif ukaz == ZAKLJUCI:\n",
    "            stanje.shrani_v_datoteko(IME_DATOTEKE)\n",
    "            print('Nasvidenje!')\n",
    "            break         \n",
    "```\n",
    "\n",
    "Da ne bomo izumljali tople vode, si bomo stanje zapisali v JSON datoteko. Ker Python našega stanja ne zna samodejno spremeniti v JSON, bomo sami napisali vmesno pretvorbo v slovar. Vse naše objekte moramo tako dopolniti z metodami za pretvorbo v slovar in iz slovarja. Prve bodo običajne metode, ki bodo vrnile slovar, na primer:\n",
    "\n",
    "```python\n",
    "class Spisek:\n",
    "    ...\n",
    "    \n",
    "    def v_slovar(self):\n",
    "        return {\n",
    "            \"ime\": self.ime,\n",
    "            \"opravila\": [opravilo.v_slovar() for opravilo in self.opravila],\n",
    "        }\n",
    "\n",
    "\n",
    "class Opravilo:\n",
    "    ...\n",
    "    \n",
    "    def v_slovar(self):\n",
    "        return {\n",
    "            \"ime\": self.ime,\n",
    "            \"opis\": self.opis,\n",
    "            \"rok\": date.isoformat(self.rok) if self.rok else None,\n",
    "            \"opravljeno\": self.opravljeno,\n",
    "        }\n",
    "```\n",
    "\n",
    "Metode za pretvorbo iz slovarja pa so malo bolj posebne, saj pred trenutkom, ko preberemo slovar, še nimamo objekta, na katerih bi jih poklicali. Lahko bi si sicer napisali funkcije kot so `preberi_opravilo_iz_slovarja` ali `preberi_spisek_iz_slovarja`, vendar te funkcije zelo naravno spadajo v ustrezne razrede. Hkrati pa niso metode, saj nimajo objekta, na katerih jih lahko pokličemo. Gre za _statične metode_, ki so v resnici navadne funkcije, le da jih kot metode zapišemo v razred in na vrhu označimo s `@staticmethod` (v resnici bi bilo boljša možnost `@classmethod`, ampak o tem kdaj drugič):\n",
    "\n",
    "```python\n",
    "class Spisek:\n",
    "    ...\n",
    "    \n",
    "    @staticmethod\n",
    "    def iz_slovarja(slovar):\n",
    "        spisek = Spisek(slovar[\"ime\"])\n",
    "        spisek.opravila = [\n",
    "            Opravilo.iz_slovarja(sl_opravila) for sl_opravila in slovar[\"opravila\"]\n",
    "        ]\n",
    "        return spisek\n",
    "\n",
    "\n",
    "class Opravilo:\n",
    "    ...\n",
    "    \n",
    "    @staticmethod\n",
    "    def iz_slovarja(slovar):\n",
    "        return Opravilo(\n",
    "            slovar[\"ime\"],\n",
    "            slovar[\"opis\"],\n",
    "            date.fromisoformat(slovar[\"rok\"]) if slovar[\"rok\"] else None,\n",
    "            slovar[\"opravljeno\"],\n",
    "        )\n",
    "```\n",
    "\n",
    "Opremljeni z vsemi funkcijami lahko shranjevanje in nalaganje stanja napišemo kot:\n",
    "\n",
    "```python\n",
    "import json\n",
    "\n",
    "class Stanje:\n",
    "    ...\n",
    "    \n",
    "    def shrani_v_datoteko(self, ime_datoteke):\n",
    "        with open(ime_datoteke, 'w') as dat:\n",
    "            slovar = self.v_slovar()\n",
    "            json.dump(slovar, dat)\n",
    "    \n",
    "    @staticmethod\n",
    "    def preberi_iz_datoteke(ime_datoteke):\n",
    "        with open(ime_datoteke) as dat:\n",
    "            slovar = json.load(dat)\n",
    "            return Stanje.iz_slovarja(slovar)"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_metadata_filter": "-all",
   "formats": "md:myst",
   "text_representation": {
    "extension": ".md",
    "format_name": "myst",
    "format_version": 0.13,
    "jupytext_version": "1.10.1"
   }
  },
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.13"
  },
  "source_map": [
   14,
   40,
   99
  ]
 },
 "nbformat": 4,
 "nbformat_minor": 5
}